// src/lib.ts
import {
  cursorCharLeft,
  cursorCharRight,
  cursorDocEnd,
  cursorDocStart,
  cursorLineBoundaryRight,
  cursorLineDown,
  cursorLineStart,
  cursorLineUp,
  cursorPageDown,
  cursorPageUp,
  deleteCharBackward,
  deleteCharForward,
  indentLess,
  indentMore,
  insertNewlineAndIndent,
  selectAll,
  selectCharLeft,
  selectCharRight,
  selectDocEnd,
  selectDocStart,
  selectGroupLeft,
  selectGroupRight,
  selectLineBoundaryRight,
  selectLineDown,
  selectLineStart,
  selectLineUp,
  selectPageDown,
  selectPageUp,
  selectParentSyntax,
  toggleComment
} from "@codemirror/commands";
import {
  EditorSelection as EditorSelection2,
  EditorState as EditorState2,
  Facet,
  Transaction
} from "@codemirror/state";
import {
  Decoration,
  EditorView as EditorView2,
  ViewPlugin,
  WidgetType,
  drawSelection,
  getPanel,
  keymap,
  showPanel
} from "@codemirror/view";
import { SearchQuery } from "@codemirror/search";
import { matchBrackets, syntaxTree } from "@codemirror/language";

// src/state.ts
import { StateEffect, StateField } from "@codemirror/state";
var modeEffect = StateEffect.define();
var modeField = StateField.define({
  create() {
    return { type: 0 /* Normal */, minor: 2 /* Normal */ };
  },
  update(mode, tr) {
    for (const effect of tr.effects) {
      if (effect.is(modeEffect)) {
        mode = effect.value;
      }
    }
    return mode;
  }
});
function sameMode(mode, otherMode) {
  return mode.type === otherMode.type && mode.minor === otherMode.minor;
}
function sameModeState(mode, otherMode) {
  return sameMode(mode, otherMode) && mode.count === otherMode.count && mode.expecting === otherMode.expecting;
}
function modeStatus(mode) {
  let result = "";
  if (mode.type === 1 /* Insert */) {
    return result;
  }
  if (mode.count) {
    result += mode.count;
  }
  result += minorModeStr(mode.minor);
  if (mode.expecting) {
    result += mode.expecting.minor;
  }
  return result;
}
function minorModeStr(minor) {
  switch (minor) {
    case 2 /* Normal */:
      return "";
    case 3 /* Goto */:
      return "g";
    case 5 /* Match */:
      return "m";
    case 6 /* Space */:
      return "<space>";
    default: {
      if (true) {
        throw new Error("Unexpected mode");
      }
    }
  }
}
var yankEffect = StateEffect.define();
var registersField = StateField.define({
  create() {
    return {};
  },
  update(registers, tr) {
    for (const effect of tr.effects) {
      if (effect.is(yankEffect)) {
        if (!Array.isArray(effect.value)) {
          return effect.value.reset;
        }
        const [reg, value] = effect.value;
        if (value.length === 0) {
          const { [reg]: _reg, ...rest } = registers;
          registers = rest;
        } else {
          registers = { ...registers, [reg]: value };
        }
      }
    }
    return registers;
  }
});
var historyEffect = StateEffect.define();
var historyField = StateField.define({
  create() {
    return { checkpoints: [], cursor: -1, pending: null, head: null };
  },
  update(value, tr) {
    for (const effect of tr.effects) {
      if (effect.is(historyEffect)) {
        const effectValue = effect.value;
        switch (effectValue.type) {
          case "move": {
            const cursor = value.cursor + effectValue.offset;
            value = {
              checkpoints: value.checkpoints,
              pending: null,
              cursor,
              head: value.head ?? effectValue.head ?? null
            };
            break;
          }
          case "add": {
            if (true) {
              if (value.pending && effectValue.temp) {
                throw new Error("Unexpected temp");
              }
            }
            const checkpoints = value.cursor === -1 ? value.checkpoints : value.checkpoints.slice(value.cursor + 1);
            value = effectValue.temp ? {
              checkpoints: value.checkpoints,
              cursor: value.cursor,
              pending: effectValue.state,
              head: value.head
            } : {
              checkpoints: [effectValue.state, ...checkpoints],
              cursor: -1,
              pending: null,
              head: null
            };
            break;
          }
          case "commit": {
            if (true) {
              if (!value.pending) {
                throw new Error("unexpected commit");
              }
            }
            if (effectValue.state.doc.eq(value.pending.doc)) {
              value = {
                ...value,
                pending: null
              };
              break;
            }
            const checkpoints = value.cursor === -1 ? value.checkpoints : value.checkpoints.slice(value.cursor + 1);
            value = {
              checkpoints: [value.pending, ...checkpoints],
              cursor: -1,
              pending: null,
              head: null
            };
            break;
          }
        }
      }
    }
    return value;
  }
});

// src/panels.ts
import { EditorView } from "@codemirror/view";
import { EditorSelection } from "@codemirror/state";
var panelStyles = EditorView.theme({
  ".cm-hx-status-panel": {
    display: "flex",
    "justify-content": "space-between",
    "font-family": "monospace"
  },
  ".cm-hx-command-panel": {
    display: "flex",
    justifyContent: "space-between",
    fontFamily: "monospace",
    minHeight: "18px"
  },
  ".cm-hx-command-input": {
    fontFamily: "monospace",
    fontSize: "inherit",
    border: "none",
    outline: "none",
    padding: "0",
    margin: "0",
    background: "inherit"
  },
  ".cm-hx-command-popup": {
    position: "fixed",
    background: "#ccc"
  },
  ".cm-hx-command-help": {
    border: "1px solid #777",
    background: "#ddd",
    padding: "2px",
    whiteSpace: "preserve"
  }
});
var CommandPanel = class {
  constructor(view, commandFacet, startSearch2) {
    this.view = view;
    this.commandFacet = commandFacet;
    this.startSearch = startSearch2;
    this.dom = $el("div");
    this.minorCommand = $el("span");
    this.inputContainer = $el("span");
    this.commandPopup = $el("div");
    this.dom.append(this.inputContainer);
    this.dom.append(this.minorCommand);
    this.dom.append(this.commandPopup);
    this.dom.classList.add("cm-hx-command-panel");
    $style(this.inputContainer, { visibility: "hidden" });
    this.label = $el("span");
    this.inputContainer.append(this.label);
    this.commandPopup.classList.add("cm-hx-command-popup");
    this.help = $el("div");
    this.autocomplete = $el("div");
    this.help.hidden = true;
    this.help.classList.add("cm-hx-command-help");
    this.autocomplete.classList.add("cm-hx-command-autocomplete");
    this.commandPopup.append(this.help);
    this.commandPopup.append(this.autocomplete);
    $style(this.minorCommand, { minWidth: "8em", textAlign: "center " });
  }
  dom;
  minorCommand;
  inputContainer;
  label;
  message = false;
  commandPopup;
  autocomplete;
  help;
  popupRequest;
  showSearchInput(global = false) {
    const input = this.searchInput(global);
    this.showInput(input, global ? "global-search:" : "search:");
  }
  showCommandInput() {
    const input = this.commandInput();
    this.showInput(input, ":");
  }
  showMinor(command) {
    this.minorCommand.textContent = modeStatus(command);
  }
  showInput(input, label) {
    this.label.textContent = label;
    $style(this.label, { color: "" });
    this.inputContainer.append(input);
    $style(this.inputContainer, { visibility: "" });
    input.focus();
  }
  createInput({
    onInput,
    onClose,
    placeholder
  }) {
    const input = $el("input");
    if (placeholder) {
      input.placeholder = placeholder;
    }
    input.classList.add("cm-hx-command-input");
    input.type = "text";
    let open = true;
    input.addEventListener("blur", () => {
      if (open) {
        onClose(false, input.value);
      }
    });
    input.addEventListener("input", () => {
      onInput(input.value);
    });
    input.addEventListener("keydown", (event) => {
      if (event.isComposing) {
        return;
      }
      const isEnter = event.key === "Enter";
      if (isEnter || event.key === "Escape") {
        open = false;
        onClose(isEnter, input.value);
      }
    });
    return input;
  }
  commandInput() {
    const { view } = this;
    const initialSelection = view.state.selection;
    const initialScroll = view.scrollSnapshot();
    const isNumber = (cmd) => /^\d+$/.test(cmd);
    return this.createInput({
      onClose: (commit, value) => {
        this.hidePopup();
        const [cmd, ...args] = value.split(/ +/);
        if (commit && cmd) {
          const commands2 = view.state.facet(this.commandFacet);
          const command = commands2.find(
            (command2) => command2.name === cmd || command2.aliases?.some((alias) => alias === cmd)
          );
          const result = command ? command.handler(view, args) : {
            message: `no such command: '${cmd}'`,
            error: true
          };
          if (result) {
            this.showMessageAndCloseInput(result);
            return;
          }
        } else if (!commit && isNumber(cmd)) {
          view.dispatch({
            selection: initialSelection
          });
          setTimeout(() => {
            view.dispatch({
              effects: initialScroll
            });
          });
        }
        this.closeInput();
      },
      onInput: (value) => {
        const cmd = value.split(/ +/).at(0);
        if (!cmd) {
          this.hidePopup();
          return;
        }
        if (isNumber(cmd)) {
          const lineNo = Number(cmd);
          if (lineNo >= 1 && lineNo <= view.state.doc.lines) {
            const line = view.state.doc.line(lineNo);
            view.dispatch({
              selection: EditorSelection.cursor(line.from),
              effects: EditorView.scrollIntoView(line.from, { y: "center" })
            });
          }
          return;
        }
        const commands2 = view.state.facet(this.commandFacet);
        const options = commands2.filter(
          (command) => command.name.startsWith(cmd) || command.aliases?.some((alias) => alias.startsWith(cmd))
        );
        if (options.length === 0) {
          this.hidePopup();
          return;
        }
        const match = options.find(
          (command) => command.name === cmd || command.aliases?.some((alias) => alias === cmd)
        );
        this.showPopup(options, match);
      }
    });
  }
  showError(message) {
    this.showMessage({ message, error: true });
  }
  showMessage(messageOrResult) {
    if (messageOrResult == null) {
      return;
    }
    const [message, error] = typeof messageOrResult === "string" ? [messageOrResult, false] : [messageOrResult.message, messageOrResult.error];
    $style(this.inputContainer, { visibility: "" });
    this.message = true;
    $style(this.label, { color: error ? "red" : "" });
    this.label.textContent = message;
  }
  showMessageAndCloseInput(result) {
    this.showMessage(result);
    this.closeInput(!result);
  }
  hasMessage() {
    return this.message;
  }
  clearMessage() {
    if (this.message) {
      this.message = false;
      this.label.textContent = "";
      $style(this.inputContainer, { visibility: "hidden" });
    }
  }
  showPopup(commands2, match) {
    this.commandPopup.hidden = false;
    this.help.hidden = !match;
    if (match) {
      this.help.textContent = `${match.help}`;
      if (match.aliases && match.aliases.length > 0) {
        this.help.textContent += `
Aliases: ${match.aliases.join(",")}`;
      }
    } else {
      this.help.textContent = "";
    }
    while (commands2.length > this.autocomplete.childNodes.length) {
      const entry = $el("span");
      $style(entry, { marginRight: "1em" });
      this.autocomplete.append(entry);
    }
    for (const [i, child] of this.autocomplete.childNodes.entries()) {
      const command = commands2[i];
      if (command) {
        child.textContent = command.name;
      } else {
        break;
      }
    }
    while (this.autocomplete.childNodes.length > commands2.length) {
      this.autocomplete.lastChild?.remove();
    }
    if (this.popupRequest == null) {
      this.popupRequest = requestAnimationFrame(() => this.positionPopup());
    }
  }
  hidePopup() {
    this.commandPopup.hidden = true;
  }
  positionPopup() {
    this.popupRequest = void 0;
    if (this.commandPopup.hidden) {
      return;
    }
    const box = this.inputContainer.getBoundingClientRect();
    $style(this.commandPopup, {
      bottom: `${window.innerHeight - box.top}px`,
      left: `${box.left}px`
    });
  }
  searchInput(global) {
    const search = this.startSearch(global);
    return this.createInput({
      placeholder: search.init,
      onClose: (commit) => {
        this.showMessageAndCloseInput(search.onClose(commit));
      },
      onInput: (value) => {
        search.onInput(value);
      }
    });
  }
  closeInput(hide = true) {
    this.inputContainer.removeChild(this.inputContainer.lastChild);
    if (hide) {
      $style(this.inputContainer, { visibility: "hidden" });
    }
    requestAnimationFrame(() => {
      this.view.focus();
    });
  }
};
function statusPanel(view) {
  const dom = $el("div");
  dom.classList.add("cm-hx-status-panel");
  const mode = $el("span");
  mode.textContent = "NOR";
  dom.append(mode);
  const pos = $el("span");
  dom.append(pos);
  function setLineCol() {
    const { line, column } = lineCol(view);
    pos.textContent = `${line}:${column}`;
  }
  setLineCol();
  return {
    dom,
    setMode(modeStr) {
      mode.textContent = modeStr;
    },
    setLineCol
  };
}
function lineCol(view) {
  const head = view.state.selection.main.head;
  const lineDesc = view.state.doc.lineAt(head);
  const line = lineDesc.number;
  const column = head - lineDesc.from + 1;
  return { line, column };
}
function $el(tag) {
  return document.createElement(tag);
}
function $style(el, styles) {
  Object.assign(el.style, styles);
}

// src/lib.ts
var MODE_EFF = {
  NORMAL: modeEffect.of({
    type: 0 /* Normal */,
    minor: 2 /* Normal */
  }),
  NORMAL_GOTO: modeEffect.of({
    type: 0 /* Normal */,
    minor: 3 /* Goto */
  }),
  NORMAL_MATCH: modeEffect.of({
    type: 0 /* Normal */,
    minor: 5 /* Match */
  }),
  NORMAL_SPACE: modeEffect.of({
    type: 0 /* Normal */,
    minor: 6 /* Space */
  }),
  SELECT: modeEffect.of({
    type: 4 /* Select */,
    minor: 2 /* Normal */
  }),
  SELECT_GOTO: modeEffect.of({
    type: 4 /* Select */,
    minor: 3 /* Goto */
  }),
  SELECT_MATCH: modeEffect.of({
    type: 4 /* Select */,
    minor: 5 /* Match */
  }),
  SELECT_SPACE: modeEffect.of({
    type: 4 /* Select */,
    minor: 6 /* Space */
  }),
  INSERT: modeEffect.of({ type: 1 /* Insert */ })
};
function moveBy(view, mode, cursor, select) {
  let count = cmdCount(mode);
  if (mode.type === 4 /* Select */) {
    for (let _i = 0; _i < count; _i++) {
      select(view);
    }
  } else {
    const selection = view.state.selection.main;
    if (selection.from !== selection.to) {
      view.dispatch({
        selection: EditorSelection2.cursor(selection.head)
      });
    }
    for (let _i = 0; _i < count; _i++) {
      cursor(view);
    }
  }
  if (mode.count) {
    view.dispatch({
      effects: resetCount(mode)
    });
  }
}
function moveDown(view, mode) {
  moveBy(view, mode, cursorLineDown, selectLineDown);
}
function moveUp(view, mode) {
  moveBy(view, mode, cursorLineUp, selectLineUp);
}
function moveLeft(view, mode) {
  moveBy(view, mode, cursorCharLeft, selectCharLeft);
}
function moveRight(view, mode) {
  moveBy(view, mode, cursorCharRight, selectCharRight);
}
function startSearch(view, global) {
  const initialScroll = view.scrollSnapshot();
  const initialSelection = view.state.selection;
  let input = "";
  let query = null;
  function reset() {
    view.dispatch({
      selection: initialSelection
    });
    resetScroll(view, initialScroll);
  }
  return {
    init: view.state.field(registersField)["/"]?.toString() ?? "",
    onInput(input_) {
      if (input !== input_) {
        input = input_;
        query = searchQuery(input);
      } else {
        return;
      }
      if (!query?.valid) {
        return;
      }
      if (global) {
        return;
      }
      let match = query.getCursor(view.state, initialSelection.main.to).next();
      if (match.done) {
        match = query.getCursor(view.state).next();
      }
      if (!match.done) {
        const selection = EditorSelection2.range(
          match.value.from,
          match.value.to
        );
        view.dispatch({
          selection,
          effects: EditorView2.scrollIntoView(selection, { y: "center" })
        });
      }
    },
    onClose(accept) {
      if (!accept) {
        reset();
        return;
      }
      if (input != null) {
        view.dispatch({
          effects: yankEffect.of(["/", input])
        });
      }
      if (global) {
        const externalCommands = view.state.facet(externalCommandsFacet);
        const query2 = input || view.state.field(registersField)["/"];
        if (query2) {
          return externalCommands.global_search?.(query2.toString());
        }
      }
    }
  };
}
function searchQuery(query) {
  return new SearchQuery({
    search: query,
    regexp: true,
    caseSensitive: /[A-Z]/.test(query)
  });
}
var searchFacet = Facet.define({
  combine(inputs) {
    const input = inputs[0];
    return input ? searchQuery(input) : void 0;
  }
});
var helixCommandBindings = {
  insert: {
    Backspace(view) {
      deleteCharBackward(view);
    },
    Delete(view) {
      deleteCharForward(view);
    },
    Enter(view) {
      insertNewlineAndIndent(view);
    },
    Escape(view) {
      view.dispatch({
        effects: [
          MODE_EFF.NORMAL,
          historyEffect.of({ type: "commit", state: view.state })
        ]
      });
    }
  },
  normal: {
    // this one is special: we let it apply to all other minor modes
    Escape(view, mode_) {
      const mode = mode_;
      if (mode.type === 0 /* Normal */ && mode.minor === 2 /* Normal */ && mode.expecting == null && mode.count == null) {
        return true;
      }
      view.dispatch({
        effects: [MODE_EFF.NORMAL]
      });
    },
    ["/"](view) {
      const panel = getCommandPanel(view);
      panel.showSearchInput();
    },
    ...Object.fromEntries(
      Array.from({ length: 10 }, (_, count) => [
        String(count),
        (view, mode) => {
          const next = mode.count != null ? mode.count * 10 + count : count;
          if (next === 0) {
            return;
          }
          view.dispatch({
            effects: modeEffect.of({ ...mode, count: next })
          });
        }
      ])
    ),
    [":"](view, mode) {
      view.dispatch({
        effects: modeEffect.of({ ...mode, count: void 0 })
      });
      getCommandPanel(view).showCommandInput();
    },
    ["y"](view) {
      const selection = view.state.selection.main;
      const range = helixSelection(selection, view.state.doc);
      view.dispatch({
        effects: yankEffect.of([
          `"`,
          view.state.doc.slice(range.from, range.to)
        ])
      });
      getCommandPanel(view).showMessage('yanked 1 selection to register "');
    },
    ["a"]: {
      checkpoint: "temp",
      command(view) {
        const selection = view.state.selection.main;
        view.dispatch({
          effects: MODE_EFF.INSERT,
          selection: EditorSelection2.cursor(selection.to)
        });
      }
    },
    ["A"]: {
      checkpoint: "temp",
      command(view) {
        const selection = view.state.selection.main;
        const end = view.state.doc.lineAt(selection.to).to;
        view.dispatch({
          effects: MODE_EFF.INSERT,
          selection: EditorSelection2.cursor(end)
        });
      }
    },
    ["I"]: {
      checkpoint: "temp",
      command(view) {
        const selection = view.state.selection.main;
        const start = view.state.doc.lineAt(selection.from).from;
        view.dispatch({
          effects: MODE_EFF.INSERT,
          selection: EditorSelection2.cursor(start)
        });
      }
    },
    ["c"]: {
      checkpoint: "temp",
      command(view) {
        const selection = view.state.selection.main;
        const range = helixSelection(selection, view.state.doc);
        view.dispatch({
          effects: [
            yankEffect.of([`"`, view.state.doc.slice(range.from, range.to)]),
            MODE_EFF.INSERT
          ],
          changes: {
            from: range.from,
            to: range.to,
            insert: ""
          }
        });
      }
    },
    ["d"]: {
      checkpoint: true,
      command(view) {
        const selection = view.state.selection.main;
        const range = helixSelection(selection, view.state.doc);
        view.dispatch({
          effects: [
            yankEffect.of([`"`, view.state.doc.slice(range.from, range.to)]),
            MODE_EFF.NORMAL
          ],
          changes: {
            from: range.from,
            to: range.to
          }
        });
      }
    },
    ["P"]: {
      checkpoint: true,
      command(view, mode) {
        const yanked = view.state.field(registersField);
        paste(view, yanked[`"`], false, cmdCount(mode));
      }
    },
    ["p"]: {
      checkpoint: true,
      command(view, mode) {
        const yanked = view.state.field(registersField);
        paste(view, yanked[`"`], true, cmdCount(mode));
      }
    },
    ["R"]: {
      checkpoint: true,
      command(view, mode) {
        const contents = view.state.field(registersField)[`"`] ?? "";
        const count = cmdCount(mode);
        const replacement = count === 1 ? contents : contents.toString().repeat(count);
        const tr = view.state.replaceSelection(replacement);
        tr.effects = MODE_EFF.NORMAL;
        view.dispatch(tr);
      }
    },
    ["r"]: {
      checkpoint: true,
      command(view, mode) {
        view.dispatch({
          effects: modeEffect.of({
            ...mode,
            expecting: {
              minor: "r",
              callback: replaceWithChar,
              metadata: view
            }
          })
        });
      }
    },
    ["w"](view, mode) {
      if (mode.type === 0 /* Normal */) {
        const current = view.state.selection.main;
        view.dispatch({
          selection: EditorSelection2.single(Math.max(current.from, current.to))
        });
      }
      for (let _i = 0; _i < cmdCount(mode); _i++) {
        selectGroupRight(view);
      }
    },
    ["b"](view, mode) {
      if (mode.type === 0 /* Normal */) {
        const current = view.state.selection.main;
        view.dispatch({
          selection: EditorSelection2.single(Math.min(current.from, current.to))
        });
      }
      for (let _i = 0; _i < cmdCount(mode); _i++) {
        selectGroupLeft(view);
      }
    },
    ["v"](view, mode) {
      view.dispatch({
        effects: mode.type === 0 /* Normal */ ? MODE_EFF.SELECT : MODE_EFF.NORMAL
      });
    },
    ["g"](view, mode) {
      const isNormal = mode.type === 0 /* Normal */;
      view.dispatch({
        effects: isNormal ? MODE_EFF.NORMAL_GOTO : MODE_EFF.SELECT_GOTO
      });
    },
    ["Space"](view, mode) {
      const isNormal = mode.type === 0 /* Normal */;
      view.dispatch({
        effects: isNormal ? MODE_EFF.NORMAL_SPACE : MODE_EFF.SELECT_SPACE
      });
    },
    ["m"](view, mode) {
      const isNormal = mode.type === 0 /* Normal */;
      view.dispatch({
        effects: isNormal ? MODE_EFF.NORMAL_MATCH : MODE_EFF.SELECT_MATCH
      });
    },
    ["i"]: {
      checkpoint: "temp",
      command(view) {
        const selection = view.state.selection.main;
        const cursor = Math.min(selection.from, selection.to);
        view.dispatch({
          effects: MODE_EFF.INSERT,
          selection: EditorSelection2.cursor(cursor)
        });
      }
    },
    ["h"](view, mode) {
      return moveLeft(view, mode);
    },
    ["j"]: moveDown,
    ["ArrowDown"]: "j",
    ["ArrowUp"]: "k",
    ["k"]: moveUp,
    ["ArrowRight"]: "l",
    ["ArrowLeft"]: "h",
    ["l"](view, mode) {
      return moveRight(view, mode);
    },
    ["%"](view) {
      return selectAll(view);
    },
    ["o"]: {
      checkpoint: "temp",
      command(view) {
        insertLineAndEdit(view, true);
      }
    },
    ["O"]: {
      checkpoint: "temp",
      command(view) {
        insertLineAndEdit(view, false);
      }
    },
    ["f"](view, mode) {
      setFindMode(view, "f", mode, {
        inclusive: true,
        reverse: false
      });
    },
    ["F"](view, mode) {
      setFindMode(view, "F", mode, {
        inclusive: true,
        reverse: true
      });
    },
    ["t"](view, mode) {
      setFindMode(view, "t", mode, {
        inclusive: false,
        reverse: false
      });
    },
    ["T"](view, mode) {
      setFindMode(view, "T", mode, {
        inclusive: false,
        reverse: true
      });
    },
    ["u"](view, mode) {
      const { checkpoints, cursor } = view.state.field(historyField);
      const nextCursor = cursor + cmdCount(mode);
      const state = checkpoints[nextCursor];
      if (!state) {
        return true;
      }
      view.dispatch({
        changes: {
          from: 0,
          to: view.state.doc.length,
          insert: state.doc
        },
        selection: state.selection,
        effects: [
          historyEffect.of({
            type: "move",
            offset: cmdCount(mode),
            head: view.state
          }),
          resetCount(mode)
        ],
        scrollIntoView: true
      });
    },
    ["U"](view, mode) {
      const { checkpoints, cursor, head } = view.state.field(historyField);
      const nextCursor = cursor - cmdCount(mode);
      const state = nextCursor === -1 ? head : checkpoints[nextCursor];
      if (!state) {
        return true;
      }
      view.dispatch({
        changes: {
          from: 0,
          to: view.state.doc.length,
          insert: state.doc
        },
        selection: state.selection,
        effects: [
          historyEffect.of({
            type: "move",
            offset: -cmdCount(mode)
          }),
          resetCount(mode)
        ],
        scrollIntoView: true
      });
    },
    ["x"](view, mode) {
      const initial = view.state.selection.main;
      const anchorLine = view.state.doc.lineAt(initial.anchor);
      const headLine = view.state.doc.lineAt(initial.head);
      const [startLine, endLine] = anchorLine.number < headLine.number ? [anchorLine, headLine] : [headLine, anchorLine];
      const ideal = EditorSelection2.range(startLine.from, endLine.to);
      let nextSel;
      const perfectLineSelection = ideal.from === initial.from && ideal.to === initial.to;
      if (perfectLineSelection || mode.count) {
        const nextLineNumber = Math.min(
          endLine.number + cmdCount(mode),
          view.state.doc.lines
        );
        const nextLine = view.state.doc.line(nextLineNumber);
        nextSel = EditorSelection2.range(startLine.from, nextLine.to);
      } else {
        nextSel = ideal;
      }
      view.dispatch({
        selection: nextSel,
        effects: resetCount(mode)
      });
      return true;
    },
    ["n"](view, mode) {
      const query = view.state.facet(searchFacet);
      if (!query?.valid) {
        if (query) {
          showSearchError(view, query);
        }
        view.dispatch({
          effects: resetCount(mode)
        });
        return true;
      }
      let cursor = query.getCursor(view.state, view.state.selection.main.to);
      let match;
      let found = false;
      let wrapped = false;
      for (let _i = 0; _i < cmdCount(mode); _i++) {
        match = cursor.next();
        found ||= !match.done;
        if (match.done) {
          cursor = query.getCursor(view.state);
          wrapped = true;
          match = cursor.next();
          found ||= !match.done;
        }
        if (!found) {
          getCommandPanel(view).showError("No more matches");
          return;
        }
      }
      view.dispatch({
        selection: EditorSelection2.range(match.value.from, match.value.to),
        scrollIntoView: true
      });
      if (wrapped) {
        getCommandPanel(view).showMessage("Wrapped around document");
      }
    },
    ["N"](view, mode) {
      const query = view.state.facet(searchFacet);
      if (!query?.valid) {
        if (query) {
          showSearchError(view, query);
        }
        return true;
      }
      const select = (match) => {
        const selection2 = EditorSelection2.range(match.from, match.to);
        view.dispatch({
          selection: selection2,
          effects: EditorView2.scrollIntoView(selection2, { y: "center" })
        });
      };
      const cursor = query.getCursor(view.state);
      const selection = view.state.selection.main;
      const count = cmdCount(mode);
      const beforeRing = new Ring(count);
      const iter = peekable(cloned(cursor));
      const beforeIter = peekingUntil(
        iter,
        (item) => item.to >= selection.from
      );
      for (const item of {
        [Symbol.iterator]() {
          return beforeIter;
        }
      }) {
        beforeRing.push(item);
      }
      if (beforeRing.length === count) {
        select(beforeRing.first);
        return;
      }
      const afterRing = new Ring(count - beforeRing.length);
      for (const item of {
        [Symbol.iterator]() {
          return iter;
        }
      }) {
        afterRing.push(item);
      }
      const total = afterRing.length + beforeRing.length;
      if (total === 0) {
        getCommandPanel(view).showError("No more matches");
        return;
      }
      getCommandPanel(view).showMessage("Wrapped around document");
      if (total === count) {
        select(afterRing.first);
        return;
      }
      const all = [...beforeRing, ...afterRing];
      const rem = count % total;
      select(all[all.length - rem - 1]);
    },
    ["Ctrl-d"](view, mode) {
      moveBy(view, mode, cursorPageDown, selectPageDown);
    },
    ["PageDown"]: "Ctrl-d",
    ["PageUp"]: "Ctrl-u",
    ["Ctrl-u"](view, mode) {
      moveBy(view, mode, cursorPageUp, selectPageUp);
    },
    [";"](view) {
      const selection = view.state.selection.main;
      view.dispatch({
        selection: EditorSelection2.cursor(selection.head)
      });
    },
    ["Alt-;"](view) {
      const selection = view.state.selection.main;
      view.dispatch({
        selection: EditorSelection2.range(selection.head, selection.anchor),
        scrollIntoView: true
      });
    },
    ["Alt-:"](view) {
      const selection = view.state.selection.main;
      view.dispatch({
        selection: EditorSelection2.range(selection.from, selection.to)
      });
    },
    ["Alt-ArrowUp"](view) {
      return selectParentSyntax(view);
    },
    ["Alt-o"]: "Alt-ArrowUp",
    ["Alt-ArrowRight"](view) {
      moveToSibling(view, true);
    },
    ["Alt-n"]: "Alt-ArrowRight",
    ["Alt-ArrowLeft"](view) {
      moveToSibling(view, false);
    },
    ["Alt-p"]: "Alt-ArrowLeft",
    ["Ctrl-c"]: {
      checkpoint: true,
      command(view) {
        toggleComment(view);
        view.original.dispatch({
          effects: MODE_EFF.NORMAL
        });
      }
    },
    [">"]: {
      checkpoint: true,
      command(view) {
        return indentMore(view);
      }
    },
    ["<"]: {
      checkpoint: true,
      command(view) {
        return indentLess(view);
      }
    },
    ["`"]: {
      checkpoint: true,
      command(view) {
        changeCase(view, false);
      }
    },
    ["Alt-`"]: {
      checkpoint: true,
      command(view) {
        changeCase(view, true);
      }
    },
    ["~"]: {
      checkpoint: true,
      command(view) {
        changeCase(view);
      }
    },
    ["*"](view) {
      const selection = helixSelection(
        view.state.selection.main,
        view.state.doc
      );
      const selected = view.state.doc.slice(selection.from, selection.to).toString();
      view.dispatch({
        effects: yankEffect.of(["/", selected])
      });
      getCommandPanel(view).showMessage(`register '/' set to '${selected}'`);
    },
    ["_"](view) {
      const selection = helixSelection(
        view.state.selection.main,
        view.state.doc
      );
      const selected = view.state.doc.slice(selection.from, selection.to).toString();
      const trimmed = selected.trim();
      if (trimmed === selected) {
        return;
      }
      const startOffset = selected.indexOf(trimmed);
      const endOffset = selected.length - trimmed.length - startOffset;
      const anchor = selection.anchor === selection.from ? selection.anchor + startOffset : selection.anchor - endOffset;
      const head = selection.head === selection.to ? selection.head - endOffset : selection.head + startOffset;
      view.dispatch({
        selection: EditorSelection2.range(anchor, head)
      });
    }
  },
  goto: {
    ["g"](view, mode) {
      const isNormal = mode.type === 0 /* Normal */;
      isNormal ? cursorDocStart(view) : selectDocStart(view);
      view.dispatch({
        effects: isNormal ? MODE_EFF.NORMAL : MODE_EFF.SELECT
      });
    },
    ["e"](view, mode) {
      const isNormal = mode.type === 0 /* Normal */;
      isNormal ? cursorDocEnd(view) : selectDocEnd(view);
      view.dispatch({
        effects: isNormal ? MODE_EFF.NORMAL : MODE_EFF.SELECT
      });
    },
    ["h"](view, mode) {
      const isNormal = mode.type === 0 /* Normal */;
      isNormal ? cursorLineStart(view) : selectLineStart(view);
      view.dispatch({
        effects: isNormal ? MODE_EFF.NORMAL : MODE_EFF.SELECT
      });
    },
    ["j"](view, mode) {
      const isNormal = mode.type === 0 /* Normal */;
      moveDown(view, mode);
      view.dispatch({
        effects: isNormal ? MODE_EFF.NORMAL : MODE_EFF.SELECT
      });
    },
    ["k"](view, mode) {
      const isNormal = mode.type === 0 /* Normal */;
      moveUp(view, mode);
      view.dispatch({
        effects: isNormal ? MODE_EFF.NORMAL : MODE_EFF.SELECT
      });
    },
    ["l"](view, mode) {
      const isNormal = mode.type === 0 /* Normal */;
      isNormal ? cursorLineBoundaryRight(view) : selectLineBoundaryRight(view);
      view.dispatch({
        effects: isNormal ? MODE_EFF.NORMAL : MODE_EFF.SELECT
      });
    },
    ["n"](view, mode) {
      const externalCommandDefs = view.state.facet(externalCommandsFacet);
      const result = externalCommandDefs[":buffer-next"]?.();
      getCommandPanel(view).showMessage(result);
      view.dispatch({
        effects: mode.type === 0 /* Normal */ ? MODE_EFF.NORMAL : MODE_EFF.SELECT
      });
    },
    ["p"](view, mode) {
      const externalCommandDefs = view.state.facet(externalCommandsFacet);
      const result = externalCommandDefs[":buffer-previous"]?.();
      getCommandPanel(view).showMessage(result);
      view.dispatch({
        effects: mode.type === 0 /* Normal */ ? MODE_EFF.NORMAL : MODE_EFF.SELECT
      });
    }
  },
  match: {
    ["s"]: {
      checkpoint: true,
      command(view, mode) {
        view.dispatch({
          effects: modeEffect.of({
            type: mode.type,
            minor: 5 /* Match */,
            expecting: {
              minor: "s",
              callback: surround,
              metadata: view
            }
          })
        });
      }
    },
    ["m"](view, mode) {
      const selection = matchBracket(view) ?? void 0;
      const isNormal = mode.type === 0 /* Normal */;
      view.dispatch({
        selection: selection?.from == null ? void 0 : isNormal ? EditorSelection2.cursor(selection.from) : EditorSelection2.range(
          view.state.selection.main.anchor,
          selection.from
        ),
        effects: isNormal ? MODE_EFF.NORMAL : MODE_EFF.SELECT,
        scrollIntoView: true
      });
    }
  },
  space: {
    ["y"](view) {
      const selection = view.state.selection.main;
      const range = helixSelection(selection, view.state.doc);
      navigator.clipboard.writeText(
        view.state.doc.slice(range.from, range.to).toString()
      );
      getCommandPanel(view).showMessage("yanked 1 selection to register +");
      view.dispatch({
        effects: MODE_EFF.NORMAL
      });
    },
    ["p"]: {
      checkpoint: true,
      command(view) {
        view.dispatch({ effects: MODE_EFF.NORMAL });
        navigator.clipboard.readText().then((yanked) => paste(view, yanked, true, 1, false));
      }
    },
    ["P"]: {
      checkpoint: true,
      command(view) {
        view.dispatch({ effects: MODE_EFF.NORMAL });
        navigator.clipboard.readText().then((yanked) => paste(view, yanked, false, 1, false));
      }
    },
    ["R"]: {
      checkpoint: true,
      command(view) {
        view.dispatch({ effects: MODE_EFF.NORMAL });
        navigator.clipboard.readText().then((yanked) => {
          const tr = view.state.replaceSelection(yanked);
          view.dispatch(tr);
        });
      }
    },
    ["f"](view, mode) {
      const externalCommandDefs = view.state.facet(externalCommandsFacet);
      const result = externalCommandDefs.file_picker?.();
      getCommandPanel(view).showMessage(result);
      view.dispatch({
        effects: mode.type === 0 /* Normal */ ? MODE_EFF.NORMAL : MODE_EFF.SELECT
      });
    },
    ["b"](view, mode) {
      const externalCommandsDef = view.state.facet(externalCommandsFacet);
      const result = externalCommandsDef.buffer_picker?.();
      getCommandPanel(view).showMessage(result);
      view.dispatch({
        effects: mode.type === 0 /* Normal */ ? MODE_EFF.NORMAL : MODE_EFF.SELECT
      });
    },
    ["/"](view, mode) {
      const enabled = view.state.facet(externalCommandsFacet).global_search != null;
      if (enabled) {
        getCommandPanel(view).showSearchInput(true);
      }
      view.dispatch({
        effects: mode.type === 0 /* Normal */ ? MODE_EFF.NORMAL : MODE_EFF.SELECT
      });
    }
  }
};
function paste(view, yanked, before, count, reset = true) {
  const range = view.state.selection.main;
  yanked ??= "";
  const spec = { from: range.to, insert: yanked.toString().repeat(count) };
  const change = view.state.changes(spec);
  view.dispatch(
    { changes: change },
    {
      selection: before ? { anchor: range.to, head: range.to + yanked.length } : { anchor: range.from, head: range.from + yanked.length },
      sequential: true
    },
    reset ? { effects: MODE_EFF.NORMAL } : {}
  );
}
function setFindMode(view, status, mode, metadata) {
  const effect = modeEffect.of({
    type: mode.type,
    minor: 2 /* Normal */,
    count: mode.count,
    expecting: {
      minor: status,
      callback: findText,
      metadata
    }
  });
  view.dispatch({ effects: effect });
}
function helixSelection(range, doc) {
  const to = Math.min(range.to + 1, doc.length);
  const forward = range.head >= range.anchor;
  const anchor = forward ? range.anchor : to;
  const head = forward ? to : range.head;
  return EditorSelection2.range(anchor, head);
}
function moveToSibling(view, forward) {
  const tree = syntaxTree(view.state);
  const selection = view.state.selection.main;
  let stack = tree.resolveStack(selection.from, 1);
  let sibling = null;
  while (true) {
    const node = stack.node;
    if (node && node.from <= selection.from && node.to >= selection.to) {
      sibling = forward ? node?.nextSibling : node?.prevSibling;
      if (sibling) {
        break;
      }
    }
    if (stack.next) {
      stack = stack.next;
    } else {
      break;
    }
  }
  if (!sibling) {
    view.dispatch({
      selection: EditorSelection2.range(0, view.state.doc.length),
      scrollIntoView: true
    });
    return;
  }
  view.dispatch({
    selection: EditorSelection2.range(sibling.from, sibling.to),
    scrollIntoView: true
  });
}
function insertLineAndEdit(view, below) {
  let from;
  let cursor;
  if (below) {
    const line = view.state.doc.lineAt(view.state.selection.main.to);
    from = line.to;
    cursor = from + 1;
  } else {
    const line = view.state.doc.lineAt(view.state.selection.main.from);
    from = line.from;
    cursor = from;
  }
  view.dispatch({
    changes: {
      from,
      insert: "\n"
    },
    selection: EditorSelection2.cursor(cursor),
    effects: MODE_EFF.INSERT
  });
}
function toCodemirrorKeymap(keybindings) {
  const allKeys = [
    ...new Set(
      Object.values(keybindings).flat().flatMap((binding) => Object.keys(binding))
    )
  ];
  function apply(def, view, mode) {
    if (typeof def === "function") {
      return def(view, mode);
    } else {
      const temp = def.checkpoint === "temp";
      return def.command(
        {
          original: view,
          dispatch(...args) {
            view.dispatch(commitToHistory(view, temp), ...args);
          },
          get state() {
            return view.state;
          }
        },
        mode
      );
    }
  }
  function getExplicitCommand(key, bindings) {
    while (true) {
      const binding = bindings[key];
      if (typeof binding === "string") {
        key = binding;
        continue;
      }
      return binding;
    }
  }
  const codemirrorKeybindings = [];
  for (const key of allKeys) {
    const insertCommand = getExplicitCommand(key, keybindings.insert);
    const normalCommand = getExplicitCommand(key, keybindings.normal);
    const gotoCommand = getExplicitCommand(key, keybindings.goto);
    const matchCommand = getExplicitCommand(key, keybindings.match);
    const spaceCommand = getExplicitCommand(key, keybindings.space);
    const esc = key === "Escape";
    const isChar = key.length === 1;
    const command = (view) => {
      const mode = view.state.field(modeField);
      if (mode.type === 1 /* Insert */) {
        if (insertCommand) {
          return insertCommand(view, void 0) ?? true;
        } else {
          return false;
        }
      }
      if (mode.expecting && isChar) {
        return false;
      }
      let result;
      if (esc || mode.minor === 2 /* Normal */ && normalCommand) {
        result = apply(normalCommand, view, mode);
      } else if (mode.minor === 3 /* Goto */ && gotoCommand) {
        result = apply(gotoCommand, view, mode);
      } else if (mode.minor === 5 /* Match */ && matchCommand) {
        result = apply(matchCommand, view, mode);
      } else if (mode.minor === 6 /* Space */ && spaceCommand) {
        result = apply(spaceCommand, view, mode);
      } else {
        return false;
      }
      return result ?? true;
    };
    codemirrorKeybindings.push({
      key,
      run: command
    });
  }
  return codemirrorKeybindings;
}
var EndLineCursor = class extends WidgetType {
  toDOM() {
    const span = document.createElement("span");
    span.classList.add("cm-hx-cursor");
    span.classList.add("cm-hx-cursor-endline");
    span.textContent = " ";
    return span;
  }
};
var cursorMark = Decoration.mark({ class: "cm-hx-cursor" });
var endlineCursorWidget = Decoration.widget({ widget: new EndLineCursor() });
function drawCursorMark(selection, doc) {
  const head = selection.main.head;
  const line = doc.lineAt(head);
  if (line.to === head) {
    return Decoration.set(
      endlineCursorWidget.range(selection.main.head, selection.main.head)
    );
  } else {
    return Decoration.set(
      cursorMark.range(selection.main.head, selection.main.head + 1)
    );
  }
}
function letThrough(tr) {
  return tr;
}
var changeFilter = EditorState2.transactionFilter.from(
  modeField,
  (mode) => mode.type === 1 /* Insert */ ? letThrough : (tr) => {
    const userEvent = tr.annotation(Transaction.userEvent);
    if (userEvent == null) {
      return tr;
    }
    if (!userEvent.startsWith("input")) {
      return tr;
    }
    if (!userEvent.startsWith("input.type")) {
      return tr;
    }
    if (mode.minor !== 2 /* Normal */) {
      return {
        effects: mode.type === 0 /* Normal */ ? MODE_EFF.NORMAL : MODE_EFF.SELECT
      };
    }
    return [];
  }
);
var inputHandler = EditorView2.inputHandler.from(
  modeField,
  (mode) => (view, _from, _to, text) => {
    if (mode.type === 1 /* Insert */) {
      return false;
    }
    if (mode.expecting) {
      mode.expecting.callback(view, text, mode.expecting.metadata);
      return true;
    }
    return false;
  }
);
var updateListener = EditorView2.updateListener.of((viewUpdate) => {
  const { state, startState } = viewUpdate;
  const panel = getPanel(viewUpdate.view, statusPanel);
  const mode = state.field(modeField);
  const startMode = startState.field(modeField);
  if (mode !== startMode) {
    const startExternalMode = toExternalMode(startMode);
    const externalMode = toExternalMode(mode);
    if (startExternalMode !== externalMode) {
      panel.setMode(externalMode);
    }
  }
  panel.setLineCol();
});
var helixKeymap = keymap.of(toCodemirrorKeymap(helixCommandBindings));
var externalCommandsFacet = Facet.define({
  combine(values) {
    const handlers = [...values];
    handlers.reverse();
    if (true) {
      const merged = values.reduce((acc, defs) => {
        for (const key of Object.keys(defs)) {
          if (acc[key] == null) {
            acc[key] = 1;
          } else {
            acc[key]++;
          }
        }
        return acc;
      }, {});
      const multiple = Object.entries(merged).flatMap(
        ([key, count]) => count > 1 ? [key] : []
      );
      if (multiple.length > 0) {
        console.warn(
          `Multiple definitions found for external commands: ${multiple.join(
            ", "
          )}`
        );
      }
    }
    return handlers.reduce((acc, def) => {
      return { ...acc, ...def };
    }, {});
  }
});
function snapshot(state, global = false) {
  return {
    registers: state.field(registersField),
    ...global ? {} : {
      history: state.field(historyField)
    }
  };
}
function globalStateSync(state) {
  return [
    {
      effects: yankEffect.of({ reset: state.field(registersField) })
    }
  ];
}
var commands = Facet.define({
  combine(commands2) {
    return commands2.flat();
  }
});
var resetMode = MODE_EFF.NORMAL;
function helix(options = {}) {
  const registerState = options.globalInit ?? options.init;
  const initialRegisters = registerState instanceof EditorState2 ? registerState.field(registersField) : registerState ? registerState.registers : void 0;
  const initialHistory = options.init instanceof EditorState2 ? options.init.field(historyField) : options.init ? options.init.history : void 0;
  return [
    EditorView2.theme({
      ".cm-hx-block-cursor .cm-cursor": {
        display: "none !important"
      },
      ".cm-hx-block-cursor .cm-hx-cursor": {
        background: "#ccc"
      },
      // WARNING: flaky
      ".cm-searchMatch": {
        background: "initial"
      }
    }),
    panelStyles,
    drawSelection({
      cursorBlinkRate: 0
    }),
    helixKeymap,
    modeField,
    initialHistory ? historyField.init(() => initialHistory) : historyField,
    initialRegisters ? registersField.init(() => initialRegisters) : registersField,
    searchFacet.from(registersField, (registers) => registers["/"]?.toString()),
    changeFilter,
    inputHandler,
    EditorView2.decorations.compute(["selection", "doc"], (state) => {
      return drawCursorMark(state.selection, state.doc);
    }),
    updateListener,
    showPanel.of(statusPanel),
    showPanel.of(commandPanel),
    ViewPlugin.define((view) => {
      const cursorShape = options?.config?.["editor.cursor-shape.insert"];
      view.scrollDOM.classList.add("cm-hx-block-cursor");
      return {
        update(update) {
          const mode = update.state.field(modeField);
          const startMode = update.startState.field(modeField);
          const panel = getCommandPanel(view);
          if (panel.hasMessage() && update.docChanged || update.selectionSet) {
            panel.clearMessage();
          }
          const modeChanged = !sameMode(mode, startMode);
          if (modeChanged || !sameModeState(mode, startMode)) {
            panel.showMinor(mode);
          }
          if (modeChanged && cursorShape === "bar") {
            if (mode.type === 1 /* Insert */) {
              view.scrollDOM.classList.remove("cm-hx-block-cursor");
            } else {
              view.scrollDOM.classList.add("cm-hx-block-cursor");
            }
          }
        }
      };
    }),
    commands.of([
      {
        name: "goto",
        aliases: ["g"],
        help: "Goto line number",
        handler(view, args) {
          if (args.length === 0) {
            return { message: "Line number required", error: true };
          }
          const lineNo = Number(args[0]);
          if (!Number.isFinite(lineNo) || lineNo <= 0) {
            return { message: "Invalid line number", error: true };
          }
          const effectiveLine = Math.min(lineNo, view.state.doc.lines);
          const line = view.state.doc.line(effectiveLine);
          view.dispatch({
            selection: EditorSelection2.cursor(line.from),
            scrollIntoView: true
          });
        }
      },
      {
        name: "clipboard-yank",
        help: "Yank main selection into system clipboard",
        handler(view) {
          const selection = view.state.selection.main;
          const range = helixSelection(selection, view.state.doc);
          navigator.clipboard.writeText(
            view.state.doc.slice(range.from, range.to).toString()
          );
          return { message: "Yanked main selection to + register" };
        }
      }
    ]),
    commands.compute([externalCommandsFacet], (state) => {
      const externalCommands = state.facet(externalCommandsFacet);
      const hardcodedCommands = [
        [":buffer-next", "Goto next buffer", ["bn", "bnext"]],
        [":buffer-previous", "Goto previous buffer", ["bp", "bprev"]],
        [":buffer-close", "Close the current buffer", ["bc", "bclose"]]
      ];
      return hardcodedCommands.filter(([name]) => !!externalCommands[name]).map(([name, help, aliases]) => ({
        name: name.slice(1),
        aliases,
        help,
        handler(view) {
          const defs = view.state.facet(externalCommandsFacet);
          return defs[name]?.();
        }
      }));
    })
  ];
}
function commandPanel(view) {
  return new CommandPanel(
    view,
    commands,
    (global) => startSearch(view, global)
  );
}
function getCommandPanel(view) {
  return getPanel(view, commandPanel);
}
function toExternalMode(mode) {
  switch (mode.type) {
    case 0 /* Normal */:
      return "NOR";
    case 4 /* Select */:
      return "SEL";
    case 1 /* Insert */:
      return "INS";
  }
}
function findText(view, text, {
  inclusive,
  reverse
}) {
  const mode = view.state.field(modeField);
  const count = mode.type === 1 /* Insert */ ? 1 : cmdCount(mode);
  const select = mode.type === 4 /* Select */;
  const selection = view.state.selection.main;
  const start = selection.head;
  const doc = view.state.doc.toString();
  let rawIndex = start;
  for (let _i = 0; _i < count; _i++) {
    if (reverse && rawIndex === 0) {
      rawIndex = -1;
      break;
    }
    rawIndex = reverse ? doc.lastIndexOf(text, rawIndex - 1) : doc.indexOf(text, rawIndex + 1);
    if (rawIndex < 0) {
      break;
    }
  }
  const resetEffect = select ? MODE_EFF.SELECT : MODE_EFF.NORMAL;
  if (rawIndex === -1) {
    view.dispatch({
      effects: resetEffect
    });
    return;
  }
  const index = inclusive ? rawIndex : reverse ? rawIndex + 1 : rawIndex - 1;
  const newSelection = select ? EditorSelection2.range(selection.anchor, index) : EditorSelection2.range(selection.head, index);
  view.dispatch({
    effects: resetEffect,
    selection: newSelection
  });
}
var PAIRS = {
  "(": ["(", ")", true],
  ")": ["(", ")", false],
  "{": ["{", "}", true],
  "}": ["{", "}", false],
  "[": ["[", "]", true],
  "]": ["[", "]", false],
  "<": ["<", ">", true],
  ">": ["<", ">", false]
};
var MATCHEABLE = /* @__PURE__ */ new Set([...Object.keys(PAIRS), `"`, "'"]);
function matchBracket(view) {
  const selection = view.state.selection.main;
  const char = view.state.doc.slice(selection.head, selection.head + 1).toString();
  if (!MATCHEABLE.has(char)) {
    return null;
  }
  const open = PAIRS[char]?.[2] ?? false;
  const match = matchBrackets(
    view.state,
    selection.head + (open ? 0 : 1),
    open ? 1 : -1
  );
  if (match) {
    return match.end;
  }
}
function surround(view, char, proxy) {
  const pair = PAIRS[char];
  const selection = view.state.selection.main;
  const range = helixSelection(selection, view.state.doc);
  const open = pair?.[0] ?? char;
  const close = pair?.[1] ?? char;
  view.dispatch({
    changes: [
      {
        from: range.from,
        insert: open
      },
      {
        from: range.to,
        insert: close
      }
    ],
    effects: MODE_EFF.NORMAL
  });
  const newSelection = proxy.original.state.selection.main;
  const offset = newSelection.anchor === newSelection.from ? 1 : -1;
  const anchor = newSelection.anchor - offset;
  const head = newSelection.head + offset;
  proxy.original.dispatch({
    selection: EditorSelection2.range(anchor, head)
  });
}
function commitToHistory(view, temp = false) {
  return {
    effects: historyEffect.of({
      type: "add",
      state: view.state,
      temp
    })
  };
}
function changeCase(view, upper) {
  const selection = view.state.selection.main;
  const selected = view.state.doc.slice(selection.from, selection.to);
  let insert;
  if (upper == null) {
    insert = [...selected.toString()].map((char) => {
      let next = char.toUpperCase();
      return next === char ? char.toLowerCase() : next;
    }).join("");
  } else if (upper) {
    insert = selected.toString().toUpperCase();
  } else {
    insert = selected.toString().toLowerCase();
  }
  view.dispatch({
    changes: {
      from: selection.from,
      to: selection.to,
      insert
    }
  });
}
function replaceWithChar(view, char, viewProxy) {
  const selection = view.state.selection.main;
  const selected = view.state.doc.slice(selection.from, selection.to);
  viewProxy.dispatch({
    changes: {
      from: selection.from,
      to: selection.to,
      insert: selected.toString().replace(/[^\n]/g, char)
    },
    effects: MODE_EFF.NORMAL
  });
}
function showSearchError(view, query) {
  let message = "";
  try {
    query.getCursor(view.state);
  } catch (error) {
    message = error?.message;
  }
  getCommandPanel(view).showError(
    `Invalid regex /${query.search}/: ${message}`
  );
}
function cmdCount(mode) {
  return mode.count ?? 1;
}
function resetCount(mode) {
  const { count: _count, ...rest } = mode;
  return modeEffect.of(rest);
}
var Ring = class {
  items;
  head;
  length;
  maxLength;
  constructor(length) {
    this.items = Array.from({ length });
    this.maxLength = length;
    this.head = 0;
    this.length = 0;
  }
  get first() {
    return this.items[this.start];
  }
  push(item) {
    this.items[this.head] = item;
    this.head += 1;
    this.length += 1;
    this.head %= this.maxLength;
    this.length = Math.min(this.maxLength, this.length);
  }
  merge(other) {
    for (let i = 0; i < other.length; i++) {
      const item = other.items[(other.start + i) % other.maxLength];
      this.push(item);
    }
  }
  *[Symbol.iterator]() {
    for (let i = 0; i < this.length; i++) {
      yield this.items[(this.start + i) % this.maxLength];
    }
  }
  get start() {
    if (this.length < this.maxLength) {
      return 0;
    } else {
      return (this.head + 1) % this.maxLength;
    }
  }
};
function peekingUntil(iter, check) {
  return {
    next() {
      const item = iter.peek();
      if (item.done) {
        return item;
      }
      if (check(item.value)) {
        return { value: void 0, done: true };
      }
      return iter.next();
    }
  };
}
function peekable(iter) {
  let next = iter.next();
  const peekIter = {
    next() {
      const item = next;
      if (!item.done) {
        next = iter.next();
      }
      return item;
    },
    peek() {
      return next;
    }
  };
  return peekIter;
}
function cloned(iter) {
  return {
    next() {
      return { ...iter.next() };
    }
  };
}
function resetScroll(view, effect) {
  requestAnimationFrame(
    () => requestAnimationFrame(() => {
      view.dispatch({ effects: effect });
    })
  );
}
export {
  commands,
  externalCommandsFacet as externalCommands,
  globalStateSync,
  helix,
  resetMode,
  snapshot
};
